﻿using HeuristicLab.Algorithms.PushGP.Interpreter;
using HeuristicLab.Algorithms.PushGP.Stack;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    public abstract class LoopExpression : StatefullExpression
    {
        protected long destinationIndex;
        protected long currentIndex;
        protected long incrementor;
        protected Expression body;

        public LoopExpression()
        { }

        /// <summary>
        /// Clone constructor
        /// </summary>
        protected LoopExpression(LoopExpression origin, Cloner cloner) : base(origin, cloner)
        {
            destinationIndex = origin.destinationIndex;
            currentIndex = origin.currentIndex;
            incrementor = origin.incrementor;
            body = origin.body;
        }

        protected void Eval(IInterpreter interpreter, IStack<Expression> sourceStack, bool pushCurrentIndex = false)
        {
            // if not initialized
            if (body == null)
            {
                if (HasInsufficientArguments(interpreter, sourceStack))
                {
                    return;
                }

                InitState(interpreter, sourceStack);

                interpreter.ExecStack.Push(this, body);

                return;
            }

            // if loop end reached
            if (destinationIndex == currentIndex)
            {
                PushLastIteration(interpreter);
                Clear();

                return;
            }

            PushIteration(interpreter);
            currentIndex += incrementor;
        }

        protected virtual void Clear()
        {
            body = null;
        }

        protected virtual void PushIteration(IInterpreter interpreter)
        {
            interpreter.IntegerStack.Push(currentIndex);
            interpreter.ExecStack.Push(this, body);
        }

        protected virtual void PushLastIteration(IInterpreter interpreter)
        {
            interpreter.IntegerStack.Push(currentIndex);
            interpreter.ExecStack.Push(body);
        }

        public override bool Equals(object obj)
        {
            if (!base.Equals(obj))
            {
                return false;
            }

            var other = obj as LoopExpression;

            var isBodyEqual =
                (body == null && other.body == null) ||
                (body != null && other.body != null && body.Equals(other.body));

            return isBodyEqual &&
                currentIndex == other.currentIndex &&
                destinationIndex == other.destinationIndex &&
                incrementor == other.incrementor;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        protected override int InitId()
        {
            unchecked
            {
                var hash = 19;

                if (body != null)
                {
                    hash *= 31 + body.GetHashCode();
                }

                hash *= 31 + currentIndex.GetHashCode();
                hash *= 31 + destinationIndex.GetHashCode();
                hash *= 31 + incrementor.GetHashCode();

                return hash;
            }
        }

        protected abstract bool HasInsufficientArguments(IInterpreter interpreter, IStack<Expression> sourceStack);
        protected abstract void InitState(IInterpreter interpreter, IStack<Expression> sourceStack);
    }
}
